/* HPCU Bi-Weekly Challenge: Bouncing Particles
 * Author: Aaron Weeden, Shodor, 2014
 */

/* Libraries */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Constants */
#define EMPTY -1

#define NORTH 0
#define SOUTH 1
#define N_Y_DIRS 2

#define EAST 0
#define WEST 1
#define N_X_DIRS 2

#define N_PARTS 20
#define N_ROWS 12
#define N_COLS 11
#define N_TIME 5

/* Globals */
int Grid[N_ROWS][N_COLS];
int Rows[N_PARTS];
int Cols[N_PARTS];
int XDirs[N_PARTS];
int YDirs[N_PARTS];

/* Function declarations */
void empty();
void randomize();
void display(int i);
void move();

/* Main */
int main() {
    int i;

    srandom(time(NULL));
    empty();
    randomize();
    display(0);
    for (i = 1; i < N_TIME; i++) {
        move();
        display(i);
    }

    return 0;
}

/* Function definitions */
void empty() {
    int row, col;

    for (row = 0; row < N_ROWS; row++) {
        for (col = 0; col < N_COLS; col++) {
            Grid[row][col] = EMPTY;
        }
    }
}

void randomize() {
    int i;
    int row, col;

    for (i = 0; i < N_PARTS; i++) {
        do {
            row = N_ROWS * random() / RAND_MAX;
            col = N_COLS * random() / RAND_MAX;
        } while (Grid[row][col] != EMPTY);

        Rows[i] = row;
        Cols[i] = col;
        XDirs[i] = N_X_DIRS * random() / RAND_MAX;
        YDirs[i] = N_Y_DIRS * random() / RAND_MAX;

        Grid[row][col] = i;
    }
}

void display(int i) {
    int row, col;

    printf("Time step %d:\n", i);
    printf("-");
    for (col = 0; col < N_COLS; col++) {
        printf("-");
    }
    printf("-\n");
    for (row = 0; row < N_ROWS; row++) {
        printf("|");
        for (col = 0; col < N_COLS; col++) {
            if (Grid[row][col] == EMPTY) {
                printf(" ");
            } else {
                printf("o");
            }
        }
        printf("|\n");
    }
    printf("-");
    for (col = 0; col < N_COLS; col++) {
        printf("-");
    }
    printf("-\n");
}

void move() {
    int i;

    for (i = 0; i < N_PARTS; i++) {
        /* particles hitting walls change direction */
        if (YDirs[i] == NORTH && Rows[i] == 0) {
            YDirs[i] = SOUTH;
        }
        if (XDirs[i] == EAST && Cols[i] == N_COLS-1) {
            XDirs[i] = WEST;
        }
        if (YDirs[i] == SOUTH && Rows[i] == N_ROWS-1) {
            YDirs[i] = NORTH;
        }
        if (XDirs[i] == WEST && Cols[i] == 0) {
            XDirs[i] = EAST;
        }
        /* particles colliding in opposite directions change direction */
        if (YDirs[i] == NORTH && Rows[i] > 0 &&
                Grid[Rows[i]-1][Cols[i]] != EMPTY &&
                YDirs[Grid[Rows[i]-1][Cols[i]]] == SOUTH) {
            YDirs[i] = SOUTH;
            YDirs[Grid[Rows[i]-1][Cols[i]]] = NORTH;
        }
        if (XDirs[i] == EAST && Cols[i] < N_COLS-1 &&
                Grid[Rows[i]][Cols[i]+1] != EMPTY &&
                XDirs[Grid[Rows[i]][Cols[i]+1]] == WEST) {
            XDirs[i] = WEST;
            XDirs[Grid[Rows[i]][Cols[i]+1]] = EAST;
        }
        if (YDirs[i] == SOUTH && Rows[i] < N_ROWS-1 &&
                Grid[Rows[i]+1][Cols[i]] != EMPTY &&
                YDirs[Grid[Rows[i]+1][Cols[i]]] == NORTH) {
            YDirs[i] = NORTH;
            YDirs[Grid[Rows[i]+1][Cols[i]]] = SOUTH;
        }
        if (XDirs[i] == WEST && Cols[i] > 0 &&
                Grid[Rows[i]][Cols[i]-1] != EMPTY &&
                XDirs[Grid[Rows[i]][Cols[i]-1]] == EAST) {
            XDirs[i] = EAST;
            XDirs[Grid[Rows[i]][Cols[i]-1]] = WEST;
        }
        /* particles move to empty squares or bounce away */
        if (YDirs[i] == NORTH && XDirs[i] == EAST && Rows[i] > 0 &&
                Cols[i] < N_COLS-1) {
            if (Grid[Rows[i]-1][Cols[i]+1] == EMPTY) {
                Grid[Rows[i]][Cols[i]] = EMPTY;
                Grid[Rows[i]-1][Cols[i]+1] = i;
                Rows[i]--;
                Cols[i]++;
            } else {
                YDirs[i] = SOUTH;
                XDirs[i] = WEST;
            }
        }
        if (YDirs[i] == NORTH && XDirs[i] == WEST && Rows[i] > 0 &&
                Cols[i] > 0) {
            if (Grid[Rows[i]-1][Cols[i]-1] == EMPTY) {
                Grid[Rows[i]][Cols[i]] = EMPTY;
                Grid[Rows[i]-1][Cols[i]-1] = i;
                Rows[i]--;
                Cols[i]--;
            } else {
                YDirs[i] = SOUTH;
                XDirs[i] = EAST;
            }
        }
        if (YDirs[i] == SOUTH && XDirs[i] == EAST && Rows[i] < N_ROWS-1 &&
                Cols[i] < N_COLS-1) {
            if (Grid[Rows[i]+1][Cols[i]+1] == EMPTY) {
                Grid[Rows[i]][Cols[i]] = EMPTY;
                Grid[Rows[i]+1][Cols[i]+1] = i;
                Rows[i]++;
                Cols[i]++;
            } else {
                YDirs[i] = NORTH;
                XDirs[i] = WEST;
            }
        }
        if (YDirs[i] == SOUTH && XDirs[i] == WEST && Rows[i] < N_ROWS-1 &&
                Cols[i] > 0) {
            if (Grid[Rows[i]+1][Cols[i]-1] == EMPTY) {
                Grid[Rows[i]][Cols[i]] = EMPTY;
                Grid[Rows[i]+1][Cols[i]-1] = i;
                Rows[i]++;
                Cols[i]--;
            } else {
                YDirs[i] = NORTH;
                XDirs[i] = EAST;
            }
        }
    }
}
